#pragma once
#include "lasp_daq.h"
#include <algorithm>
#include <cassert>
#include <chrono>
#include <iostream>
#include <stdexcept>
#include <thread>
#include <uldaq.h>
#include <vector>

using std::atomic;
using std::cerr;
using std::endl;
using rte = std::runtime_error;

class DT9837A : public Daq {

  DaqDeviceHandle _handle = 0;
  std::mutex _daqmutex;

  std::thread _thread;
  atomic<bool> _stopThread{false};
  atomic<StreamStatus> _streamStatus;

  const us _nFramesPerBlock;

  void threadFcn(InDaqCallback inCallback, OutDaqCallback outcallback);

public:
  DaqDeviceHandle getHandle() const { return _handle; }
  /**
   * @brief Create a DT9837A instance.
   *
   * @param devinfo DeviceInfo to connect to
   * @param config DaqConfiguration settings
   */
  DT9837A(const DeviceInfo &devinfo, const DaqConfiguration &config);

  virtual ~DT9837A();

  bool isRunning() const;

  void stop() override final;

  friend class InBufHandler;
  friend class OutBufHandler;

  virtual void start(InDaqCallback inCallback,
                     OutDaqCallback outCallback) override final;

  virtual StreamStatus getStreamStatus() const override {
    return _streamStatus;
  }
};

/**
 * @brief Helper class for managing input and output samples of the DAQ device.
 */
class BufHandler {
protected:
  /**
   * @brief Reference to underlying Daq
   */
  DT9837A &daq;
  /**
   * @brief The type of data, in this case always double precision floats
   */
  const DataTypeDescriptor dtype_descr = dtype_desc_fl64;
  /**
   * @brief The number of channels, number of frames per callback (block).
   */
  us nchannels, nFramesPerBlock;
  /**
   * @brief Sampling frequency in Hz
   */
  double samplerate;
  std::vector<double> buf;
  /**
   * @brief Whether the top / bottom part of the buffer are ready to be
   * enqueued
   */
  bool topenqueued, botenqueued;

  /**
   * @brief Counter for the total number of frames acquired / sent since the
   * start of the stream.
   */
  us totalFramesCount = 0;
  long long buffer_mid_idx;

public:
  /**
   * @brief Initialize bufhandler
   *
   * @param daq
   * @param nchannels
   */
  BufHandler(DT9837A &daq, const us nchannels)
      : daq(daq), dtype_descr(daq.dtypeDescr()), nchannels(nchannels),
        nFramesPerBlock(daq.framesPerBlock()), samplerate(daq.samplerate()),
        buf(2 * nchannels *
                nFramesPerBlock, // Watch the two here, the top and the bottom!
            0),
        buffer_mid_idx(nchannels * nFramesPerBlock) {
    assert(nchannels > 0);
  }
};
/**
 * @brief Specific helper for the input buffer
 */
class InBufHandler : public BufHandler {
  bool monitorOutput;
  InDaqCallback cb;

public:
  InBufHandler(DT9837A &daq, InDaqCallback cb);
  void start();
  /**
   * @brief InBufHandler::operator()()
   *
   * @return true on success
   */
  bool operator()();
  ~InBufHandler();
  
};
class OutBufHandler : public BufHandler {
  OutDaqCallback cb;

public:
  OutBufHandler(DT9837A &daq, OutDaqCallback cb);
  void start();
  /**
   * @brief OutBufHandler::operator()
   *
   * @return true on success
   */
  bool operator()();

  ~OutBufHandler();
};
