/* #define DEBUGTRACE_ENABLED */
#include "lasp_threadedindatahandler.h"
#include "debugtrace.hpp"
#include "lasp_thread.h"
#include <future>
#include <thread>
#include <queue>
#include <optional>

using namespace std::literals::chrono_literals;
using lck = std::scoped_lock<std::mutex>;
using rte = std::runtime_error;
using std::cerr;
using std::endl;

class SafeQueue {
  std::queue<DaqData> _queue;
  std::mutex _mtx;
  std::atomic_int32_t _contents {0};
  public:
  void push(const DaqData& d) {
    DEBUGTRACE_ENTER;
    lck lock(_mtx);
    _queue.push(d);
    _contents++;
  }
  DaqData pop() {
    DEBUGTRACE_ENTER;
    if (empty()) {
      throw rte("BUG: Pop on empty queue");
    }
    lck lock(_mtx);

    /* DaqData d(std::move(_queue.front())); */
    DaqData d(_queue.front());
    _queue.pop();
    _contents--;

    return d;
  }
  /**
   * @brief Empty implemented using atomic var, safes some mutex lock/unlock
   * cycles.
   *
   * @return true if queue is empty
   */
  bool empty() const { return _contents == 0; }
};

ThreadedInDataHandler::ThreadedInDataHandler(StreamMgr &mgr)
  : InDataHandler(mgr), _queue(std::make_unique<SafeQueue>()) {

    DEBUGTRACE_ENTER;

    // Initialize thread pool, if not already done
    getPool();
  }

bool ThreadedInDataHandler::inCallback(const DaqData &daqdata) {
  DEBUGTRACE_ENTER;

  if (!_lastCallbackResult) {
    return false;
  }

  _queue->push(daqdata);

  if (!_thread_running && (!_stopThread) && _lastCallbackResult) {
    auto &pool = getPool();
    DEBUGTRACE_PRINT("Pushing new thread in pool");
    _thread_running = true;
    pool.push_task(&ThreadedInDataHandler::threadFcn, this);
  }

  return _lastCallbackResult;
}

ThreadedInDataHandler::~ThreadedInDataHandler() {

  DEBUGTRACE_ENTER;
  _stopThread = true;

  // Then wait in steps for the thread to stop running.
  while (_thread_running) {
    std::this_thread::sleep_for(10us);
  }
}

void ThreadedInDataHandler::threadFcn() {

  DEBUGTRACE_ENTER;

  while(!_queue->empty() && !_stopThread) {

    // Call inCallback_threaded
    if (!inCallback_threaded(_queue->pop())) {
      cerr << "*********** Callback result returned false! *************"
        << endl;
      _lastCallbackResult = false;
    }
  }
  _thread_running = false;
}
